#pragma once

#include "../mem/mem.h"
#include "authmap.h"
#include "buddy.h"
#include "commitrange.h"
#include "commonconfig.h"
#include "defaultpagemapentry.h"
#include "empty_range.h"
#include "globalrange.h"
#include "indirectrange.h"
#include "largebuddyrange.h"
#include "logrange.h"
#include "noprange.h"
#include "pagemap.h"
#include "pagemapregisterrange.h"
#include "palrange.h"
#include "range_helpers.h"
#include "smallbuddyrange.h"
#include "staticconditionalrange.h"
#include "statsrange.h"
#include "subrange.h"
